# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
# https://github.com/google-research/disentanglement_lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGES:
# - converted from tensorflow to pytorch
# - removed gin config
# - uses disent objects and classes
# - renamed functions

"""
Implementation of Disentanglement, Completeness and Informativeness.
Based on "A Framework for the Quantitative Evaluation of Disentangled
Representations" (https://openreview.net/forum?id=By-7dz-AZ).
"""

import logging

import numpy as np
import scipy
import scipy.stats
from tqdm import tqdm

from disent.dataset import DisentDataset
from disent.metrics import utils
from disent.metrics.utils import make_metric

log = logging.getLogger(__name__)


# ========================================================================= #
# dci                                                                       #
# ========================================================================= #


@make_metric("dci", fast_kwargs=dict(num_train=1000, num_test=500))
def metric_dci(
    dataset: DisentDataset,
    representation_function: callable,
    num_train: int = 10000,
    num_test: int = 5000,
    batch_size: int = 16,
    boost_mode="sklearn",
    show_progress=False,
):
    """Computes the DCI scores according to Sec 2.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and
        outputs a dim_representation sized representation for each observation.
      num_train: Number of points used for training.
      num_test: Number of points used for testing.
      batch_size: Batch size for sampling.
      boost_mode: which boosting algorithm should be used [sklearn, xgboost, lightgbm] (this can have a significant effect on score)
      show_progress: If a tqdm progress bar should be shown
    Returns:
      Dictionary with average disentanglement score, completeness and
        informativeness (train and test).
    """
    log.debug("Generating training set.")
    # mus_train are of shape [num_codes, num_train], while ys_train are of shape
    # [num_factors, num_train].
    mus_train, ys_train = utils.generate_batch_factor_code(
        dataset, representation_function, num_train, batch_size, show_progress=False
    )
    assert mus_train.shape[1] == num_train
    assert ys_train.shape[1] == num_train
    mus_test, ys_test = utils.generate_batch_factor_code(
        dataset, representation_function, num_test, batch_size, show_progress=False
    )

    log.debug("Computing DCI metric.")
    scores = _compute_dci(mus_train, ys_train, mus_test, ys_test, boost_mode=boost_mode, show_progress=show_progress)

    return scores


def _compute_dci(mus_train, ys_train, mus_test, ys_test, boost_mode="sklearn", show_progress=False):
    """Computes score based on both training and testing codes and factors."""
    importance_matrix, train_err, test_err = _compute_importance_gbt(
        mus_train, ys_train, mus_test, ys_test, boost_mode=boost_mode, show_progress=show_progress
    )
    assert importance_matrix.shape[0] == mus_train.shape[0]
    assert importance_matrix.shape[1] == ys_train.shape[0]
    return {
        "dci.informativeness_train": train_err,  # "dci.explicitness" -- Measuring Disentanglement: A Review of Metrics
        "dci.informativeness_test": test_err,  # "dci.explicitness" -- Measuring Disentanglement: A Review of Metrics
        "dci.disentanglement": _disentanglement(
            importance_matrix
        ),  # "dci.modularity"   -- Measuring Disentanglement: A Review of Metrics
        "dci.completeness": _completeness(
            importance_matrix
        ),  # "dci.compactness"  -- Measuring Disentanglement: A Review of Metrics
    }


def _compute_importance_gbt(x_train, y_train, x_test, y_test, boost_mode="sklearn", show_progress=False):
    """Compute importance based on gradient boosted trees."""
    num_factors = y_train.shape[0]
    num_codes = x_train.shape[0]
    importance_matrix = np.zeros(shape=[num_codes, num_factors], dtype=np.float64)
    train_loss = []
    test_loss = []
    for i in tqdm(range(num_factors), disable=(not show_progress)):
        if boost_mode == "sklearn":
            from sklearn.ensemble import GradientBoostingClassifier

            model = GradientBoostingClassifier()
        elif boost_mode == "xgboost":
            from xgboost import XGBClassifier

            model = XGBClassifier()
        elif boost_mode == "lightgbm":
            from lightgbm import LGBMClassifier

            model = LGBMClassifier()
        else:
            raise KeyError(f"Invalid boosting mode: {boost_mode=}")

        model.fit(x_train.T, y_train[i, :])
        importance_matrix[:, i] = np.abs(model.feature_importances_)
        train_loss.append(np.mean(model.predict(x_train.T) == y_train[i, :]))
        test_loss.append(np.mean(model.predict(x_test.T) == y_test[i, :]))

    return importance_matrix, np.mean(train_loss), np.mean(test_loss)


def _disentanglement_per_code(importance_matrix):
    """Compute disentanglement score of each code."""
    # importance_matrix is of shape [num_codes, num_factors].
    return 1.0 - scipy.stats.entropy(importance_matrix.T + 1e-11, base=importance_matrix.shape[1])


def _disentanglement(importance_matrix):
    """Compute the disentanglement score of the representation."""
    per_code = _disentanglement_per_code(importance_matrix)
    if importance_matrix.sum() == 0.0:
        importance_matrix = np.ones_like(importance_matrix)
    code_importance = importance_matrix.sum(axis=1) / importance_matrix.sum()
    return np.sum(per_code * code_importance)


def _completeness_per_factor(importance_matrix):
    """Compute completeness of each factor."""
    # importance_matrix is of shape [num_codes, num_factors].
    return 1.0 - scipy.stats.entropy(importance_matrix + 1e-11, base=importance_matrix.shape[0])


def _completeness(importance_matrix):
    """ "Compute completeness of the representation."""
    per_factor = _completeness_per_factor(importance_matrix)
    if importance_matrix.sum() == 0.0:
        importance_matrix = np.ones_like(importance_matrix)
    factor_importance = importance_matrix.sum(axis=0) / importance_matrix.sum()
    return np.sum(per_factor * factor_importance)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

"""
The disent registry only contains implemented versions of
classes and functions, no interfaces are included.
  - for interfaces and creating your own data types, extend
    the base classes from the various locations.

*NB* All modules and classes are lazily imported!

You can register your own modules and classes using the provided decorator:
eg. `DATASET.register(...options...)(your_function_or_class)`
"""

# from disent.registry._registry import ProvidedValue
# from disent.registry._registry import StaticImport
# from disent.registry._registry import DictProviders
# from disent.registry._registry import RegexProvidersSearch

from disent.registry._registry import StaticValue
from disent.registry._registry import LazyValue
from disent.registry._registry import LazyImport
from disent.registry._registry import Registry
from disent.registry._registry import RegistryImports
from disent.registry._registry import RegexConstructor
from disent.registry._registry import RegexRegistry


# ========================================================================= #
# DATASETS - should be synchronized with: `disent/dataset/data/__init__.py` #
# ========================================================================= #


# TODO: this is not yet used in disent.data or disent.frameworks
DATASETS: RegistryImports["torch.utils.data.Dataset"] = RegistryImports("DATASETS")
# groundtruth -- impl
DATASETS["cars3d_x128"] = LazyImport("disent.dataset.data._groundtruth__cars3d.Cars3dData")
DATASETS["cars3d"] = LazyImport("disent.dataset.data._groundtruth__cars3d.Cars3d64Data")
DATASETS["dsprites"] = LazyImport("disent.dataset.data._groundtruth__dsprites.DSpritesData")
DATASETS["mpi3d_toy"] = LazyImport("disent.dataset.data._groundtruth__mpi3d.Mpi3dData", subset="toy")
DATASETS["mpi3d_realistic"] = LazyImport("disent.dataset.data._groundtruth__mpi3d.Mpi3dData", subset="realistic")
DATASETS["mpi3d_real"] = LazyImport("disent.dataset.data._groundtruth__mpi3d.Mpi3dData", subset="real")
DATASETS["smallnorb_x96"] = LazyImport("disent.dataset.data._groundtruth__norb.SmallNorbData")
DATASETS["smallnorb"] = LazyImport("disent.dataset.data._groundtruth__norb.SmallNorb64Data")
DATASETS["shapes3d"] = LazyImport("disent.dataset.data._groundtruth__shapes3d.Shapes3dData")
# groundtruth -- impl synthetic
DATASETS["xcolumns"] = LazyImport("disent.dataset.data._groundtruth__xcolumns.XColumnsData")
DATASETS["xyobject"] = LazyImport("disent.dataset.data._groundtruth__xyobject.XYObjectData")
DATASETS["xyobject_shaded"] = LazyImport("disent.dataset.data._groundtruth__xyobject.XYObjectShadedData")
DATASETS["xysquares"] = LazyImport("disent.dataset.data._groundtruth__xysquares.XYSquaresData")
DATASETS["xysquares_minimal"] = LazyImport("disent.dataset.data._groundtruth__xysquares.XYSquaresMinimalData")
DATASETS["xysinglesquare"] = LazyImport("disent.dataset.data._groundtruth__xysquares.XYSingleSquareData")


# ========================================================================= #
# SAMPLERS - should be synchronized with:                                   #
#            `disent/dataset/sampling/__init__.py`                          #
# ========================================================================= #


# TODO: this is not yet used in disent.data or disent.frameworks
# changes here should also update
SAMPLERS: RegistryImports["disent.dataset.sampling.BaseDisentSampler"] = RegistryImports("SAMPLERS")
# [ground truth samplers]
SAMPLERS["gt_dist"] = LazyImport("disent.dataset.sampling._groundtruth__dist.GroundTruthDistSampler")
SAMPLERS["gt_pair"] = LazyImport("disent.dataset.sampling._groundtruth__pair.GroundTruthPairSampler")
SAMPLERS["gt_pair_orig"] = LazyImport("disent.dataset.sampling._groundtruth__pair_orig.GroundTruthPairOrigSampler")
SAMPLERS["gt_single"] = LazyImport("disent.dataset.sampling._groundtruth__single.GroundTruthSingleSampler")
SAMPLERS["gt_triple"] = LazyImport("disent.dataset.sampling._groundtruth__triplet.GroundTruthTripleSampler")
# [any dataset samplers]
SAMPLERS["single"] = LazyImport("disent.dataset.sampling._single.SingleSampler")
SAMPLERS["random"] = LazyImport("disent.dataset.sampling._random__any.RandomSampler")
# [episode samplers]
SAMPLERS["random_episode"] = LazyImport("disent.dataset.sampling._random__episodes.RandomEpisodeSampler")


# ========================================================================= #
# FRAMEWORKS - should be synchronized with:                                 #
#             `disent/frameworks/ae/__init__.py`                            #
#             `disent/frameworks/ae/experimental/__init__.py`               #
#             `disent/frameworks/vae/__init__.py`                           #
#             `disent/frameworks/vae/experimental/__init__.py`              #
# ========================================================================= #


# TODO: this is not yet used in disent.frameworks
FRAMEWORKS: RegistryImports["disent.frameworks.DisentFramework"] = RegistryImports("FRAMEWORKS")
# [AE]
FRAMEWORKS["tae"] = LazyImport("disent.frameworks.ae._supervised__tae.TripletAe")
FRAMEWORKS["ae"] = LazyImport("disent.frameworks.ae._unsupervised__ae.Ae")
FRAMEWORKS["ada_ae"] = LazyImport("disent.frameworks.ae._weaklysupervised__adaae.AdaAe")  # ae version of ada_vae
FRAMEWORKS["adaneg_tae"] = LazyImport(
    "disent.frameworks.ae._supervised__adaneg_tae.AdaNegTripletAe"
)  # ae version of adaneg_tvae
FRAMEWORKS["adaneg_tae_d"] = LazyImport(
    "disent.frameworks.ae._unsupervised__dotae.DataOverlapTripletAe"
)  # ae version of adaneg_tvae_d
# [VAE]
FRAMEWORKS["tvae"] = LazyImport("disent.frameworks.vae._supervised__tvae.TripletVae")
FRAMEWORKS["betatc_vae"] = LazyImport("disent.frameworks.vae._unsupervised__betatcvae.BetaTcVae")
FRAMEWORKS["beta_vae"] = LazyImport("disent.frameworks.vae._unsupervised__betavae.BetaVae")
FRAMEWORKS["dfc_vae"] = LazyImport("disent.frameworks.vae._unsupervised__dfcvae.DfcVae")
FRAMEWORKS["dip_vae"] = LazyImport("disent.frameworks.vae._unsupervised__dipvae.DipVae")
FRAMEWORKS["info_vae"] = LazyImport("disent.frameworks.vae._unsupervised__infovae.InfoVae")
FRAMEWORKS["vae"] = LazyImport("disent.frameworks.vae._unsupervised__vae.Vae")
FRAMEWORKS["ada_vae"] = LazyImport("disent.frameworks.vae._weaklysupervised__adavae.AdaVae")
FRAMEWORKS["adaneg_tvae"] = LazyImport("disent.frameworks.vae._supervised__adaneg_tvae.AdaNegTripletVae")
FRAMEWORKS["adaneg_tvae_d"] = LazyImport("disent.frameworks.vae._unsupervised__dotvae.DataOverlapTripletVae")


# ========================================================================= #
# RECON_LOSSES - should be synchronized with:                               #
#                `disent/frameworks/helper/reconstructions.py`              #
# ========================================================================= #


RECON_LOSSES: RegexRegistry["disent.frameworks.helper.reconstructions.ReconLossHandler"] = RegexRegistry(
    "RECON_LOSSES"
)  # TODO: we need a regex version of RegistryImports
# [STANDARD LOSSES]
RECON_LOSSES["mse"] = LazyImport(
    "disent.frameworks.helper.reconstructions.ReconLossHandlerMse"
)  # from the normal distribution - real values in the range [0, 1]
RECON_LOSSES["mae"] = LazyImport("disent.frameworks.helper.reconstructions.ReconLossHandlerMae")  # mean absolute error
# [STANDARD DISTRIBUTIONS]
RECON_LOSSES["bce"] = LazyImport(
    "disent.frameworks.helper.reconstructions.ReconLossHandlerBce"
)  # from the bernoulli distribution - binary values in the set {0, 1}
RECON_LOSSES["bernoulli"] = LazyImport(
    "disent.frameworks.helper.reconstructions.ReconLossHandlerBernoulli"
)  # reduces to bce - binary values in the set {0, 1}
RECON_LOSSES["c_bernoulli"] = LazyImport(
    "disent.frameworks.helper.reconstructions.ReconLossHandlerContinuousBernoulli"
)  # bernoulli with a computed offset to handle values in the range [0, 1]
RECON_LOSSES["normal"] = LazyImport(
    "disent.frameworks.helper.reconstructions.ReconLossHandlerNormal"
)  # handle all real values
# [REGEX LOSSES]
RECON_LOSSES.register_regex(
    pattern=r"^([a-z\d]+)_([a-z\d]+_[a-z\d]+)_l(\d+\.\d+)_k(\d+\.\d+)_norm_([a-z]+)$",
    example="mse_xy8_abs63_l1.0_k1.0_norm_none",
    factory_fn="disent.frameworks.helper.reconstructions._make_aug_recon_loss_l_w_n",
)
RECON_LOSSES.register_regex(
    pattern=r"^([a-z\d]+)_([a-z\d]+_[a-z\d]+)_norm_([a-z]+)$",
    example="mse_xy8_abs63_norm_none",
    factory_fn="disent.frameworks.helper.reconstructions._make_aug_recon_loss_l1_w1_n",
)
RECON_LOSSES.register_regex(
    pattern=r"^([a-z\d]+)_([a-z\d]+_[a-z\d]+)$",
    example="mse_xy8_abs63",
    factory_fn="disent.frameworks.helper.reconstructions._make_aug_recon_loss_l1_w1_nnone",
)


# ========================================================================= #
# LATENT_HANDLERS - should be synchronized with:                            #
#                  `disent/frameworks/helper/latent_distributions.py`       #
# ========================================================================= #


LATENT_HANDLERS: RegistryImports["disent.frameworks.helper.latent_distributions.LatentDistsHandler"] = RegistryImports(
    "LATENT_HANDLERS"
)
LATENT_HANDLERS["normal"] = LazyImport("disent.frameworks.helper.latent_distributions.LatentDistsHandlerNormal")
LATENT_HANDLERS["laplace"] = LazyImport("disent.frameworks.helper.latent_distributions.LatentDistsHandlerLaplace")


# ========================================================================= #
# OPTIMIZER                                                                 #
# ========================================================================= #


# default learning rate for each optimizer
_LR = 1e-3

OPTIMIZERS: RegistryImports["torch.optim.Optimizer"] = RegistryImports("OPTIMIZERS")
# [torch]
OPTIMIZERS["adadelta"] = LazyImport(lr=_LR, import_path="torch.optim.adadelta.Adadelta")
OPTIMIZERS["adagrad"] = LazyImport(lr=_LR, import_path="torch.optim.adagrad.Adagrad")
OPTIMIZERS["adam"] = LazyImport(lr=_LR, import_path="torch.optim.adam.Adam")
OPTIMIZERS["adamax"] = LazyImport(lr=_LR, import_path="torch.optim.adamax.Adamax")
OPTIMIZERS["adam_w"] = LazyImport(lr=_LR, import_path="torch.optim.adamw.AdamW")
OPTIMIZERS["asgd"] = LazyImport(lr=_LR, import_path="torch.optim.asgd.ASGD")
OPTIMIZERS["lbfgs"] = LazyImport(lr=_LR, import_path="torch.optim.lbfgs.LBFGS")
OPTIMIZERS["rmsprop"] = LazyImport(lr=_LR, import_path="torch.optim.rmsprop.RMSprop")
OPTIMIZERS["rprop"] = LazyImport(lr=_LR, import_path="torch.optim.rprop.Rprop")
OPTIMIZERS["sgd"] = LazyImport(lr=_LR, import_path="torch.optim.sgd.SGD")
OPTIMIZERS["sparse_adam"] = LazyImport(lr=_LR, import_path="torch.optim.sparse_adam.SparseAdam")
# [torch_optimizer]
OPTIMIZERS["acc_sgd"] = LazyImport(lr=_LR, import_path="torch_optimizer.AccSGD")
OPTIMIZERS["ada_bound"] = LazyImport(lr=_LR, import_path="torch_optimizer.AdaBound")
OPTIMIZERS["ada_mod"] = LazyImport(lr=_LR, import_path="torch_optimizer.AdaMod")
OPTIMIZERS["adam_p"] = LazyImport(lr=_LR, import_path="torch_optimizer.AdamP")
OPTIMIZERS["agg_mo"] = LazyImport(lr=_LR, import_path="torch_optimizer.AggMo")
OPTIMIZERS["diff_grad"] = LazyImport(lr=_LR, import_path="torch_optimizer.DiffGrad")
OPTIMIZERS["lamb"] = LazyImport(lr=_LR, import_path="torch_optimizer.Lamb")
# 'torch_optimizer.Lookahead' is skipped because it is wrapped
OPTIMIZERS["novograd"] = LazyImport(lr=_LR, import_path="torch_optimizer.NovoGrad")
OPTIMIZERS["pid"] = LazyImport(lr=_LR, import_path="torch_optimizer.PID")
OPTIMIZERS["qh_adam"] = LazyImport(lr=_LR, import_path="torch_optimizer.QHAdam")
OPTIMIZERS["qhm"] = LazyImport(lr=_LR, import_path="torch_optimizer.QHM")
OPTIMIZERS["radam"] = LazyImport(lr=_LR, import_path="torch_optimizer.RAdam")
OPTIMIZERS["ranger"] = LazyImport(lr=_LR, import_path="torch_optimizer.Ranger")
OPTIMIZERS["ranger_qh"] = LazyImport(lr=_LR, import_path="torch_optimizer.RangerQH")
OPTIMIZERS["ranger_va"] = LazyImport(lr=_LR, import_path="torch_optimizer.RangerVA")
OPTIMIZERS["sgd_w"] = LazyImport(lr=_LR, import_path="torch_optimizer.SGDW")
OPTIMIZERS["sgd_p"] = LazyImport(lr=_LR, import_path="torch_optimizer.SGDP")
OPTIMIZERS["shampoo"] = LazyImport(lr=_LR, import_path="torch_optimizer.Shampoo")
OPTIMIZERS["yogi"] = LazyImport(lr=_LR, import_path="torch_optimizer.Yogi")


# ========================================================================= #
# METRIC - should be synchronized with: `disent/metrics/__init__.py`        #
# ========================================================================= #


# TODO: this is not yet used in disent.util.lightning.callbacks or disent.metrics
METRICS: RegistryImports["disent.metrics.utils._Metric"] = RegistryImports("METRICS")
METRICS["dci"] = LazyImport("disent.metrics._dci.metric_dci")
METRICS["factor_vae"] = LazyImport("disent.metrics._factor_vae.metric_factor_vae")
METRICS["mig"] = LazyImport("disent.metrics._mig.metric_mig")
METRICS["sap"] = LazyImport("disent.metrics._sap.metric_sap")
METRICS["unsupervised"] = LazyImport("disent.metrics._unsupervised.metric_unsupervised")
# register metrics
METRICS["flatness"] = LazyImport("disent.metrics._flatness.metric_flatness")
METRICS["factored_components"] = LazyImport("disent.metrics._factored_components.metric_factored_components")
METRICS["distances"] = LazyImport("disent.metrics._factored_components.metric_distances")
METRICS["linearity"] = LazyImport("disent.metrics._factored_components.metric_linearity")


# ========================================================================= #
# SCHEDULE - should be synchronized with: `disent/schedule/__init__.py`     #
# ========================================================================= #


# TODO: this is not yet used in disent.framework or disent.schedule
SCHEDULES: RegistryImports["disent.schedule.Schedule"] = RegistryImports("SCHEDULES")
SCHEDULES["clip"] = LazyImport("disent.schedule._schedule.ClipSchedule")
SCHEDULES["cosine_wave"] = LazyImport("disent.schedule._schedule.CosineWaveSchedule")
SCHEDULES["cyclic"] = LazyImport("disent.schedule._schedule.CyclicSchedule")
SCHEDULES["linear"] = LazyImport("disent.schedule._schedule.LinearSchedule")
SCHEDULES["noop"] = LazyImport("disent.schedule._schedule.NoopSchedule")


# ========================================================================= #
# MODEL - should be synchronized with: `disent/model/ae/__init__.py`        #
# ========================================================================= #


# TODO: this is not yet used in disent.framework or disent.model
MODELS: RegistryImports["disent.model._base.DisentLatentsModule"] = RegistryImports("MODELS")
# [DECODER]
MODELS["encoder_conv64"] = LazyImport("disent.model.ae._vae_conv64.EncoderConv64")
MODELS["encoder_conv64norm"] = LazyImport("disent.model.ae._norm_conv64.EncoderConv64Norm")
MODELS["encoder_fc"] = LazyImport("disent.model.ae._vae_fc.EncoderFC")
MODELS["encoder_linear"] = LazyImport("disent.model.ae._linear.EncoderLinear")
# [ENCODER]
MODELS["decoder_conv64"] = LazyImport("disent.model.ae._vae_conv64.DecoderConv64")
MODELS["decoder_conv64norm"] = LazyImport("disent.model.ae._norm_conv64.DecoderConv64Norm")
MODELS["decoder_fc"] = LazyImport("disent.model.ae._vae_fc.DecoderFC")
MODELS["decoder_linear"] = LazyImport("disent.model.ae._linear.DecoderLinear")


# ========================================================================= #
# HELPER registries                                                         #
# ========================================================================= #


# TODO: add norm support with regex?
KERNELS: RegexRegistry["torch.Tensor"] = RegexRegistry("KERNELS")
KERNELS.register_regex(
    pattern=r"^box_r(\d+)$", example="box_r31", factory_fn="disent.dataset.transform._augment._make_box_kernel"
)
KERNELS.register_regex(
    pattern=r"^gau_r(\d+)$", example="gau_r31", factory_fn="disent.dataset.transform._augment._make_gaussian_kernel"
)


# ========================================================================= #
# Registry of all Registries                                                #
# ========================================================================= #


# registry of registries
REGISTRIES: Registry[Registry] = Registry("REGISTRIES")
REGISTRIES["DATASETS"] = StaticValue(DATASETS)
REGISTRIES["SAMPLERS"] = StaticValue(SAMPLERS)
REGISTRIES["FRAMEWORKS"] = StaticValue(FRAMEWORKS)
REGISTRIES["RECON_LOSSES"] = StaticValue(RECON_LOSSES)
REGISTRIES["LATENT_HANDLERS"] = StaticValue(LATENT_HANDLERS)
REGISTRIES["OPTIMIZERS"] = StaticValue(OPTIMIZERS)
REGISTRIES["METRICS"] = StaticValue(METRICS)
REGISTRIES["SCHEDULES"] = StaticValue(SCHEDULES)
REGISTRIES["MODELS"] = StaticValue(MODELS)
REGISTRIES["KERNELS"] = StaticValue(KERNELS)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
import shutil
from tempfile import TemporaryDirectory
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np

from disent.dataset.data._groundtruth import NumpyFileGroundTruthData
from disent.dataset.util.datafile import DataFileHashed
from disent.dataset.util.datafile import DataFileHashedDlGen
from disent.util.inout.paths import modify_name_keep_ext

log = logging.getLogger(__name__)


# ========================================================================= #
# cars 3d data processing                                                   #
# ========================================================================= #


def load_cars3d_folder(raw_data_dir):
    """
    nips2015-analogy-data.tar.gz contains:
        1. /data/cars
            - list.txt: [ordered list of mat files "car_***_mesh" without the extension]
            - car_***_mesh.mat: [MATLAB file with keys: "im" (128, 128, 3, 24, 4), "mask" (128, 128, 24, 4)]
        2. /data/sprites
        3. /data/shapes48.mat
    """
    from scipy.io import loadmat

    # load image paths
    with open(os.path.join(raw_data_dir, "cars/list.txt"), "r") as img_names:
        img_paths = [os.path.join(raw_data_dir, f"cars/{name.strip()}.mat") for name in img_names.readlines()]
    # load images
    images = np.stack([loadmat(img_path)["im"] for img_path in img_paths], axis=0)
    # check size
    assert images.shape == (183, 128, 128, 3, 24, 4)
    # reshape & transpose: (183, 128, 128, 3, 24, 4) -> (4, 24, 183, 128, 128, 3) -> (17568, 128, 128, 3)
    return images.transpose([5, 4, 0, 1, 2, 3]).reshape([-1, 128, 128, 3])


def resave_cars3d_archive(orig_zipped_file, new_save_file, overwrite=False):
    """
    Convert a cars3d archive 'nips2015-analogy-data.tar.gz' to a numpy file,
    uncompressing the contents of the archive into a temporary directory in the same folder.
    """
    with TemporaryDirectory(prefix="raw_cars3d_", dir=os.path.dirname(orig_zipped_file)) as temp_dir:
        # extract zipfile and get path
        log.info(f"Extracting into temporary directory: {temp_dir}")
        shutil.unpack_archive(filename=orig_zipped_file, extract_dir=temp_dir)
        # load images
        images = load_cars3d_folder(raw_data_dir=os.path.join(temp_dir, "data"))
    # save the array
    from disent.dataset.util.formats.npz import save_dataset_array

    save_dataset_array(images, new_save_file, overwrite=overwrite, save_key="images")


def resave_cars3d_resized(orig_converted_file: str, new_resized_file: str, overwrite=False, size: int = 64):
    # load the array
    cars3d_array = np.load(orig_converted_file)["images"]
    assert cars3d_array.shape == (17568, 128, 128, 3)
    # save the array
    from disent.dataset.util.formats.npz import save_resized_dataset_array

    save_resized_dataset_array(cars3d_array, new_resized_file, overwrite=overwrite, size=size, save_key="images")


# ========================================================================= #
# cars3d data object                                                        #
# ========================================================================= #


class DataFileCars3d(DataFileHashedDlGen):
    """
    download the cars3d dataset and convert it to a numpy file.
    """

    def _generate(self, inp_file: str, out_file: str):
        resave_cars3d_archive(orig_zipped_file=inp_file, new_save_file=out_file, overwrite=True)


class DataFileCars3dResized(DataFileHashed):
    def __init__(
        self,
        cars3d_datafile: DataFileCars3d,
        # - convert file name
        out_hash: Optional[Union[str, Dict[str, str]]],
        out_name: Optional[str] = None,
        out_size: int = 64,
        # - hash settings
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        self._out_size = out_size
        self._cars3dfile = cars3d_datafile
        super().__init__(
            file_name=modify_name_keep_ext(self._cars3dfile.out_name, suffix=f"_x{out_size}")
            if (out_name is None)
            else out_name,
            file_hash=out_hash,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )

    def _prepare(self, out_dir: str, out_file: str):
        log.debug("Preparing Orig Cars3d Data:")
        cars3d_path = self._cars3dfile.prepare(out_dir)
        log.debug("Generating Resized Cars3d Data:")
        resave_cars3d_resized(
            orig_converted_file=cars3d_path, new_resized_file=out_file, overwrite=True, size=self._out_size
        )


# ========================================================================= #
# dataset_cars3d                                                            #
# ========================================================================= #


class Cars3dData(NumpyFileGroundTruthData):
    """
    Cars3D Dataset
    - Deep Visual Analogy-Making (https://papers.nips.cc/paper/5845-deep-visual-analogy-making)
      http://www.scottreed.info

    # reference implementation: https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/data/ground_truth/cars3d.py
    """

    name = "cars3d"

    factor_names = ("elevation", "azimuth", "object_type")
    factor_sizes = (4, 24, 183)  # TOTAL: 17568
    img_shape = (128, 128, 3)

    datafile = DataFileCars3d(
        uri="http://www.scottreed.info/files/nips2015-analogy-data.tar.gz",
        uri_hash={"fast": "fe77d39e3fa9d77c31df2262660c2a67", "full": "4e866a7919c1beedf53964e6f7a23686"},
        file_name="cars3d.npz",
        file_hash={"fast": "204ecb6852216e333f1b022903f9d012", "full": "46ad66acf277897f0404e522460ba7e5"},
        hash_mode="fast",
    )

    # override
    data_key = "images"


# TODO: this is very slow compared to other datasets for some reason!
#       - in memory benchmark are equivalent, eg. against Shapes3D, but when we run the
#         experiment/run.py with this its about twice as slow? Why is this?
class Cars3d64Data(Cars3dData):
    """
    Optimized version of Cars3dOrigData, that has already been re-sized to 64x64
    - This can improve run times dramatically!
    """

    img_shape = (64, 64, 3)

    datafile = DataFileCars3dResized(
        cars3d_datafile=Cars3dData.datafile,
        out_name="cars3d_x64.npz",
        out_hash={"fast": "5a85246b6f555bc6e3576ee62bf6d19e", "full": "2b900b3c5de6cd9b5df87bfc02f01f03"},
        hash_mode="fast",
        out_size=64,
    )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


if __name__ == "__main__":

    def main():
        import torch
        from tqdm import tqdm

        from disent.dataset.transform import ToImgTensorF32

        logging.basicConfig(level=logging.DEBUG)

        # original dataset
        data_128 = Cars3dData(prepare=True, transform=ToImgTensorF32(size=64))
        for i in tqdm(data_128, desc="cars3d_x128 -> 64"):
            pass
        # resized dataset
        data_64 = Cars3d64Data(prepare=True, transform=ToImgTensorF32(size=64))
        for i in tqdm(data_64, desc="cars3d_x64"):
            pass
        # check equivalence
        for obs_128, obs_64 in tqdm(zip(data_128, data_64), desc="equivalence"):
            assert torch.allclose(obs_128, obs_64)

    main()



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from dataclasses import dataclass
from typing import Sequence
from typing import Type
from typing import Union

import torch

# ========================================================================= #
# Triplet Modifications                                                     #
# ========================================================================= #


def triplet_loss(anc, pos, neg, margin_min=None, margin_max=1.0, p=1):
    """
    Standard Triplet Loss
    """
    return dist_triplet_loss(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p)


def dist_triplet_loss(pos_delta, neg_delta, margin_min=None, margin_max=1.0, p=1):
    """
    Standard Triplet Loss
    """
    if margin_min is not None:
        warnings.warn("triplet_loss does not support margin_min")
    p_dist = torch.norm(pos_delta, p=p, dim=-1)
    n_dist = torch.norm(neg_delta, p=p, dim=-1)
    loss = torch.clamp_min(p_dist - n_dist + margin_max, 0)
    return loss.mean()


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #


def triplet_sigmoid_loss(anc, pos, neg, margin_min=None, margin_max=1.0, p=1):
    """
    Sigmoid Triplet Loss
    https://arxiv.org/pdf/2003.14021.pdf
    """
    return dist_triplet_sigmoid_loss(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p)


def dist_triplet_sigmoid_loss(pos_delta, neg_delta, margin_min=None, margin_max=1.0, p=1):
    """
    Sigmoid Triplet Loss
    https://arxiv.org/pdf/2003.14021.pdf
    """
    if margin_min is not None:
        warnings.warn("triplet_sigmoid_loss does not support margin_min")
    p_dist = torch.norm(pos_delta, p=p, dim=-1)
    n_dist = torch.norm(neg_delta, p=p, dim=-1)
    loss = torch.sigmoid((1 / margin_max) * (p_dist - n_dist))
    return loss.mean()


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #


def triplet_soft_loss(anc, pos, neg, margin_min=None, margin_max=None, p=1):
    """
    Triplet Loss With Soft-Margin
    https://arxiv.org/pdf/1703.07737.pdf
    """
    return dist_triplet_soft_loss(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p)


def dist_triplet_soft_loss(pos_delta, neg_delta, margin_min=None, margin_max=None, p=1):
    """
    Triplet Loss With Soft-Margin
    https://arxiv.org/pdf/1703.07737.pdf
    """
    if margin_min is not None:
        warnings.warn("triplet_soft_loss does not support margin_min")
    if margin_max is not None:
        warnings.warn("triplet_soft_loss does not support margin_max")
    p_dist = torch.norm(pos_delta, p=p, dim=-1)
    n_dist = torch.norm(neg_delta, p=p, dim=-1)
    loss = torch.log(1 + torch.exp(p_dist - n_dist))
    return loss.mean()


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #


# def elem_triplet_loss(anc, pos, neg, margin_min=None, margin_max=1., p=1):
#     """
#     Element-Wise Triplet Loss
#     TODO: THIS SHOULD NOT WORK AT ALL! JUST TRYING SOMETHING
#     """
#     return dist_elem_triplet_loss(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p)


# def dist_elem_triplet_loss(pos_delta, neg_delta, margin_min=None, margin_max=1., p=1):
#     """
#     Element-Wise Triplet Loss
#     TODO: THIS SHOULD NOT WORK AT ALL! JUST TRYING SOMETHING
#     """
#     if margin_min is not None:
#         warnings.warn('elem_triplet_loss does not support margin_min')
#     if p != 1:
#         warnings.warn('elem_triplet_loss only supported p=1')
#     p_dist = torch.abs(pos_delta)
#     n_dist = torch.abs(neg_delta)
#     loss = torch.clamp_min(p_dist - n_dist + margin_max, 0)
#     return loss.mean()


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #


# def min_margin_triplet_loss(anc, pos, neg, margin_min=0.01, margin_max=1., p=1):
#     """
#     Min Margin Triplet Loss
#     # TODO: this is wrong?
#     """
#     return dist_min_margin_triplet_loss(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p)


# def dist_min_margin_triplet_loss(pos_delta, neg_delta, margin_min=0.01, margin_max=1., p=1):
#     """
#     Min Margin Triplet Loss
#     # TODO: this is wrong?
#     """
#     p_dist = torch.norm(pos_delta + margin_min, p=p, dim=-1)
#     n_dist = torch.norm(neg_delta, p=p, dim=-1)
#     loss = torch.clamp_min(p_dist - n_dist + margin_max, 0)
#     return loss.mean()


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #


def min_clamped_triplet_loss(anc, pos, neg, margin_min=0.01, margin_max=1.0, p=1):
    """
    Min Margin Triplet Loss
    TODO: is this better, or clamped_triplet_loss?
    TODO: could take idea from soft-margin to make this continuously differentiable?
    """
    return dist_min_clamped_triplet_loss(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p)


def dist_min_clamped_triplet_loss(pos_delta, neg_delta, margin_min=0.01, margin_max=1.0, p=1):
    """
    Min Margin Triplet Loss
    TODO: is this better, or dist_clamped_triplet_loss?
    TODO: could take idea from soft-margin to make this continuously differentiable?
    """
    p_dist = torch.norm(pos_delta, p=p, dim=-1)
    n_dist = torch.norm(neg_delta, p=p, dim=-1)
    loss = torch.clamp_min(torch.clamp_min(p_dist, margin_min) - n_dist + margin_max, 0)
    return loss.mean()


# -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #


def split_clamped_triplet_loss(anc, pos, neg, margin_min=0.01, margin_max=1.0, p=1):
    """
    Min Margin Triplet Loss
    TODO: is this better, or min_clamp_triplet_loss?
    TODO: could take idea from soft-margin to make this continuously differentiable?
    """
    return dist_split_clamped_triplet_loss(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p)


def dist_split_clamped_triplet_loss(pos_delta, neg_delta, margin_min=0.01, margin_max=1.0, p=1):
    """
    Min Margin Triplet Loss
    TODO: is this better, or dist_min_clamp_triplet_loss?
    TODO: could take idea from soft-margin to make this continuously differentiable?
    """
    p_dist = torch.norm(pos_delta, p=p, dim=-1)
    n_dist = torch.norm(neg_delta, p=p, dim=-1)
    loss = torch.clamp_min(p_dist, margin_min) - torch.clamp_max(n_dist, margin_max) + (margin_max - margin_min)
    return loss.mean()


# ========================================================================= #
# EXTRA TRIPLET LOSSES                                                      #
# ========================================================================= #


# def triplet_lossless(anc, pos, neg, margin_min=None, margin_max=3., p=1, epsilon=1e-8):
#     """
#     TODO: inputs should be from sigmoid output layer.
#     https://towardsdatascience.com/lossless-triplet-loss-7e932f990b24
#     """
#     return dist_triplet_lossless(anc - pos, anc - neg, margin_min=margin_min, margin_max=margin_max, p=p, epsilon=epsilon)


# def dist_triplet_lossless(pos_delta, neg_delta, margin_min=None, margin_max=3., p=1, epsilon=1e-8):
#     """
#     TODO: inputs should be from sigmoid output layer.
#     https://towardsdatascience.com/lossless-triplet-loss-7e932f990b24
#     """
#     if margin_min is not None:
#         warnings.warn('triplet_loss does not support margin_min')
#     p_dist = torch.norm(pos_delta, p=p, dim=-1)
#     n_dist = torch.norm(neg_delta, p=p, dim=-1)
#     # rename values
#     N = margin_max
#     # recommended value
#     beta = N
#     # non-linear activation
#     p_dist = - torch.log(-((    p_dist) / beta) + (1 + epsilon))
#     n_dist = - torch.log(-((N - n_dist) / beta) + (1 + epsilon))
#     # compute mean
#     return (p_dist + n_dist).mean()


# ========================================================================= #
# Get Triplet                                                               #
# ========================================================================= #


@dataclass
class TripletLossConfig(object):
    triplet_loss: str = "triplet"
    triplet_margin_min: float = 0.1
    triplet_margin_max: float = 10
    triplet_scale: float = 100
    triplet_p: float = 2


_TRIPLET_LOSSES = {
    "triplet": triplet_loss,
    "triplet_sigmoid": triplet_sigmoid_loss,
    "triplet_soft": triplet_soft_loss,
    # 'elem_triplet': elem_triplet_loss,
    # 'min_margin_triplet': min_margin_triplet_loss,
    "min_clamped_triplet": min_clamped_triplet_loss,
    "split_clamped_triplet": split_clamped_triplet_loss,
    # 'triplet_lossless': triplet_lossless,
}


_DIST_TRIPLET_LOSSES = {
    "triplet": dist_triplet_loss,
    "triplet_sigmoid": dist_triplet_sigmoid_loss,
    "triplet_soft": dist_triplet_soft_loss,
    # 'elem_triplet': dist_elem_triplet_loss,
    # 'min_margin_triplet': dist_min_margin_triplet_loss,
    "min_clamped_triplet": dist_min_clamped_triplet_loss,
    "split_clamped_triplet": dist_split_clamped_triplet_loss,
    # 'triplet_lossless': dist_triplet_lossless,
}


TripletConfigTypeHint = Union[TripletLossConfig, Type[TripletLossConfig]]


def configured_triplet(anc, pos, neg, cfg: TripletConfigTypeHint):
    return (
        _TRIPLET_LOSSES[cfg.triplet_loss](
            anc,
            pos,
            neg,
            margin_min=cfg.triplet_margin_min,
            margin_max=cfg.triplet_margin_max,
            p=cfg.triplet_p,
        )
        * cfg.triplet_scale
    )


def configured_dist_triplet(pos_delta, neg_delta, cfg: TripletConfigTypeHint):
    return (
        _DIST_TRIPLET_LOSSES[cfg.triplet_loss](
            pos_delta,
            neg_delta,
            margin_min=cfg.triplet_margin_min,
            margin_max=cfg.triplet_margin_max,
            p=cfg.triplet_p,
        )
        * cfg.triplet_scale
    )


def compute_triplet_loss(zs: Sequence[torch.Tensor], cfg: TripletConfigTypeHint):
    anc, pos, neg = zs
    # loss is scaled and everything
    loss = configured_triplet(anc, pos, neg, cfg=cfg)
    # return loss & log
    return loss, {f"{cfg.triplet_loss}_L{cfg.triplet_p}": loss}


def compute_dist_triplet_loss(zs_deltas: Sequence[torch.Tensor], cfg: TripletConfigTypeHint):
    pos_delta, neg_delta = zs_deltas
    # loss is scaled and everything
    loss = configured_dist_triplet(pos_delta, neg_delta, cfg=cfg)
    # return loss & log
    return loss, {f"{cfg.triplet_loss}_L{cfg.triplet_p}": loss}


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import logging
from typing import final

import numpy as np
import torch

from disent.nn.modules import DisentModule

log = logging.getLogger(__name__)


# ========================================================================= #
# Custom Base Module Involving Inputs & Representations                     #
# ========================================================================= #


# TODO: vae models should be implemented here!
#       models should potentially output distributions!
#       - but how do we handle mixmatch of stuff in frameworks?
#       - can the framework detect if the output is a distribution or not and handle accordingly?


class DisentLatentsModule(DisentModule):
    def __init__(self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1):
        super().__init__()
        self._x_shape = tuple(x_shape)
        self._x_size = int(np.prod(x_shape))
        self._z_size = int(z_size)
        self._z_multiplier = int(z_multiplier)

    def forward(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def x_shape(self):
        return self._x_shape

    @property
    def x_size(self):
        return self._x_size

    @property
    def z_size(self):
        return self._z_size

    @property
    def z_multiplier(self):
        return self._z_multiplier

    @property
    def z_total(self):
        return self._z_size * self._z_multiplier


# ========================================================================= #
# Base Encoder & Base Decoder                                               #
# ========================================================================= #


class DisentEncoder(DisentLatentsModule):
    @final
    def forward(self, x, chunk=True) -> torch.Tensor:
        """same as self.encode but with size checks"""
        # checks
        assert x.ndim == 4, f"ndim mismatch: 4 (required) != {x.ndim} (given) [shape={x.shape}]"
        assert x.shape[1:] == self.x_shape, f"x_shape mismatch: {self.x_shape} (required) != {x.shape[1:]} (batch)"
        # encode | p(z|x)
        # for a gaussian encoder, we treat z as concat(z_mean, z_logvar) where z_mean.shape == z_logvar.shape
        # ie. the first half of z is z_mean, the second half of z is z_logvar
        z = self.encode(x)
        # checks
        assert z.shape == (x.size(0), self.z_total)
        if chunk:
            z = z.split(self.z_size, dim=-1)
            assert all(s.size(1) == self.z_size for s in z)
        return z

    def encode(self, x) -> torch.Tensor:
        raise NotImplementedError


class DisentDecoder(DisentLatentsModule):
    def __init__(self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1):
        assert z_multiplier == 1, "decoder does not support z_multiplier != 1"
        super().__init__(x_shape=x_shape, z_size=z_size, z_multiplier=z_multiplier)

    @final
    def forward(self, z):
        """same as self.decode but with size checks"""
        # checks
        assert z.ndim == 2
        assert z.size(1) == self.z_size
        # decode | p(x|z)
        x_recon = self.decode(z)
        # return
        assert x_recon.shape == (z.size(0), *self.x_shape)
        return x_recon

    def decode(self, z) -> torch.Tensor:
        raise NotImplementedError


# ========================================================================= #
# Auto-Encoder Wrapper                                                      #
# ========================================================================= #


class AutoEncoder(DisentLatentsModule):
    def __init__(self, encoder: DisentEncoder, decoder: DisentDecoder):
        assert isinstance(encoder, DisentEncoder)
        assert isinstance(decoder, DisentDecoder)
        # check sizes
        assert encoder.x_shape == decoder.x_shape, "x_shape mismatch"
        assert encoder.x_size == decoder.x_size, "x_size mismatch - this should never happen if x_shape matches"
        assert encoder.z_size == decoder.z_size, "z_size mismatch"
        # initialise
        super().__init__(x_shape=decoder.x_shape, z_size=decoder.z_size, z_multiplier=encoder.z_multiplier)
        # assign
        self._encoder = encoder
        self._decoder = decoder

    def forward(self, x):
        raise RuntimeError(f"{self.__class__.__name__}.forward(...) has been disabled")

    def encode(self, x, chunk=False):
        z_raw = self._encoder(x, chunk=chunk)
        return z_raw

    def decode(self, z: torch.Tensor) -> torch.Tensor:
        """
        decode the given representation.
        the returned tensor does not have an activation applied to it!
        """
        return self._decoder(z)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Any
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union

import numpy as np
import torch
import torchvision.transforms.functional as F_tv
from PIL.Image import Image

# ========================================================================= #
# Types                                                                     #
# ========================================================================= #


_T = TypeVar("_T")

Obs = Union[np.ndarray, Image]

SizeType = Union[int, Tuple[int, int]]


# ========================================================================= #
# Functional Transforms                                                     #
# ========================================================================= #


def noop(obs: _T) -> _T:
    """
    Transform that does absolutely nothing!
    """
    return obs


def check_tensor(
    obs: Any,
    low: Optional[float] = 0.0,
    high: Optional[float] = 1.0,
    dtype: torch.dtype = torch.float32,
) -> torch.Tensor:
    """
    Check that the input is a tensor, its datatype matches, and
    that it is in the required range.
    """
    # check is a tensor
    assert torch.is_tensor(obs), "observation is not a tensor"
    # check type
    if dtype is not None:
        assert obs.dtype == dtype, f"tensor type {obs.dtype} is not required type {dtype}"
    # check range
    if low is not None:
        assert low <= obs.min(), f"minimum value of tensor {obs.min()} is less than allowed minimum value: {low}"
    if high is not None:
        assert obs.max() <= high, f"maximum value of tensor {obs.max()} is greater than allowed maximum value: {high}"
    # DONE!
    return obs


# ========================================================================= #
# Normalized Image Tensors                                                  #
# ========================================================================= #


def _is_size_different(obs: Obs, size: SizeType):
    h, w = (size, size) if isinstance(size, int) else size
    H, W = (obs.height, obs.width) if isinstance(obs, Image) else obs.shape[:2]
    return (H != h) or (W != w)


def to_img_tensor_u8(
    obs: Obs,
    size: Optional[SizeType] = None,
) -> torch.Tensor:
    """
    Basic transform that makes sure the image tensor is uint8 and a specified size.

    Steps:
    1. resize image if size is specified
    2. add missing channel to greyscale image
    3. move channels to first dim (H, W, C) -> (C, H, W)
    """
    # resize image
    if (size is not None) and _is_size_different(obs, size):
        if not isinstance(obs, Image):
            obs = F_tv.to_pil_image(obs)
        obs = F_tv.resize(obs, size=size)
    # to numpy
    if isinstance(obs, Image):
        obs = np.array(obs)
    # add missing axis
    if obs.ndim == 2:
        obs = obs[:, :, None]
    # to tensor & move axis
    obs = torch.from_numpy(obs)
    obs = torch.moveaxis(obs, -1, -3)
    # checks
    assert obs.ndim == 3
    assert obs.dtype == torch.uint8
    # done!
    return obs


def to_img_tensor_f32(
    obs: Obs,
    size: Optional[SizeType] = None,
    mean: Optional[Sequence[float]] = None,
    std: Optional[Sequence[float]] = None,
) -> torch.Tensor:
    """
    Basic transform that should be applied to most datasets, making sure
    the image tensor is float32 and a specified size.

    Steps:
        1. resize image if size is specified
        2. if we have integer inputs, divide by 255
        3. add missing channel to greyscale image
        4. move channels to first dim (H, W, C) -> (C, H, W)
        5. normalize using mean and std, values might thus be outside of the range [0, 1]
    """
    # resize image
    if (size is not None) and _is_size_different(obs, size):
        if not isinstance(obs, Image):
            obs = F_tv.to_pil_image(obs)
        obs = F_tv.resize(obs, size=size)
    # transform to tensor, add missing dims & move channel dim to front
    # TODO: this should be replaced with custom logic, this is quite slow...
    #       - benchmarks show that doing conversions as numpy first, and then using torch.from_numpy is faster!
    #       - eg. SmallNorb64Data
    #             `to_tensor(item)                                                          # 27547.81it/s
    #             `torch.from_numpy(np.transpose(item.astype('float32') / 255, [2, 0, 1]))  # 53987.90it/s
    #             `torch.from_numpy((item.astype('float32') / 255).transpose([2, 0, 1]))    # 66544.00it/s
    #             `torch.from_numpy(item.astype('float32').transpose([2, 0, 1]) / 255)      # 66511.62it/s
    #             `torch.from_numpy(item.transpose([2, 0, 1]).astype('float32') / 255)      # 66133.27it/s
    #       - eg. Cars3d64Data
    #             `to_tensor(item)                                                          # 13810.46it/s
    #             `torch.from_numpy(np.transpose(item.astype('float32') / 255, [2, 0, 1]))  # 32258.03it/s
    #             `torch.from_numpy((item.astype('float32') / 255).transpose([2, 0, 1]))    # 37861.46it/s
    #             `torch.from_numpy(item.astype('float32').transpose([2, 0, 1]) / 255)      # 33034.21it/s
    #             `torch.from_numpy(item.transpose([2, 0, 1]).astype('float32') / 255)      # 32883.32it/s
    #       - INVESTIGATE: if transpose is used, and then from_numpy is called, that references the original memory? It
    #            might then be slower to convolve this data? Speed benefits could be negated? A copy might be better?
    obs = F_tv.to_tensor(obs)
    # checks
    assert obs.ndim == 3, f"obs has does not have 3 dimensions, got: {obs.ndim} for shape: {obs.shape}"
    assert obs.dtype == torch.float32, f"obs is not dtype torch.float32, got: {obs.dtype}"
    # apply mean and std, we obs is of the shape (C, H, W)
    if (mean is not None) or (std is not None):
        obs = F_tv.normalize(obs, mean=0.0 if (mean is None) else mean, std=1.0 if (std is None) else std, inplace=True)
        assert (
            obs.dtype == torch.float32
        ), f"after normalization, tensor should remain as dtype torch.float32, got: {obs.dtype}"
    # done!
    return obs


# ========================================================================= #
# Custom Normalized Image - Faster Than Above                               #
# ========================================================================= #


# def to_img_tensor_f32(
#     x: Obs,
#     size: Optional[SizeType] = None,
#     channel_to_front: bool = None,
# ):
#     """
#     Basic transform that should be applied to
#     any dataset before augmentation.
#
#     1. resize if size is specified
#     2. if needed convert integers to float32 by dividing by 255
#     3. normalize using mean and std, values might thus be outside of the range [0, 1]
#
#     Convert PIL or uint8 inputs to float32
#     - input images should always have 2 (H, W) or 3 channels (H, W, C)
#     - output image always has size (C, H, W) with channels moved to the first dim
#     """
#     return _to_img_tensor(x, size=size, channel_to_front=channel_to_front, to_float32=True)
#
#
# def to_img_tensor_u8(
#     x: Obs,
#     size: Optional[SizeType] = None,
#     channel_to_front: bool = None,
# ):
#     """
#     Convert PIL or uint8 inputs to float32
#     - input images should always have 2 (H, W) or 3 channels (H, W, C)
#     - output image always has size (C, H, W) with channels moved to the first dim
#     """
#     return _to_img_tensor(x, size=size, channel_to_front=channel_to_front, to_float32=False)
#
#
# def _to_img_tensor(
#     x: Obs,
#     size: Optional[SizeType] = None,
#     channel_to_front: bool = None,
#     to_float32: bool = True,
# ) -> torch.Tensor:
#     assert isinstance(x, (np.ndarray, Image)), f'input is not an numpy.ndarray or PIL.Image.Image, got: {type(x)}'
#     # optionally resize the image, returns a numpy array or a PIL.Image.Image
#     x = _resize_if_needed(x, size=size)
#     # convert image to numpy
#     if isinstance(x, Image):
#         x = np.array(x)
#     # make sure 2D becomes 3D
#     if x.ndim == 2:
#         x = x[:, :, None]
#     assert x.ndim == 3, f'obs has invalid number of dimensions, required 2 or 3, got: {x.ndim} for shape: {x.shape}'
#     # convert to float32 if int or uint
#     if to_float32:
#         if x.dtype.kind in ('i', 'u'):
#             x = x.astype('float32') / 255  # faster than with torch
#     # convert to torch.Tensor and move channels (H, W, C) -> (C, H, W)
#     x = torch.from_numpy(x)
#     if channel_to_front or (channel_to_front is None):
#         x = torch.moveaxis(x, -1, 0)  # faster than the numpy version
#     # final check
#     if to_float32:
#         assert x.dtype == torch.float32, f'obs dtype invalid, required: {torch.float32}, got: {x.dtype}'
#     else:
#         assert x.dtype == torch.uint8, f'obs dtype invalid, required: {torch.uint8}, got: {x.dtype}'
#     # done
#     return x
#
#
# # ========================================================================= #
# # Resize Image Helper                                                       #
# # ========================================================================= #
#
#
# _PIL_INTERPOLATE_MODES = {
#     'nearest': 0,
#     'lanczos': 1,
#     'bilinear': 2,
#     'bicubic': 3,
#     'box': 4,
#     'hamming': 5,
# }
#
#
# def _resize_if_needed(img: Union[np.ndarray, Image], size: Optional[Union[Tuple[int, int], int]] = None) -> Union[np.ndarray, Image]:
#     # skip resizing
#     if size is None:
#         return img
#     # normalize size
#     if isinstance(size, int):
#         size = (size, size)
#     # get current image size
#     if isinstance(img, Image):
#         in_size = (img.height, img.width)
#     else:
#         assert img.ndim in (2, 3), f'image must have 2 or 3 dims, got shape: {img.shape}'
#         in_size = img.shape[:2]
#     # skip if the same size
#     if in_size == size:
#         return img
#     # normalize the image
#     if not isinstance(img, Image):
#         assert img.dtype == 'uint8'
#         # normalize image
#         if img.ndim == 3:
#             c = img.shape[-1]
#             assert c in (1, 3), f'image channel dim must be of size 1 or 3, got shape: {img.shape}'
#             img, mode = (img, 'RGB') if (c == 3) else (img[:, :, 0], 'L')
#         else:
#             mode = 'L'
#         # convert
#         img = PIL.Image.fromarray(img, mode=mode)
#     # resize
#     return img.resize(size, resample=_PIL_INTERPOLATE_MODES['bilinear'])


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
import pytest
import torch

from disent.model import AutoEncoder
from disent.model import DisentDecoder
from disent.model import DisentEncoder
from disent.model.ae import *


@pytest.mark.parametrize(
    ["encoder_cls", "decoder_cls"],
    [
        [EncoderConv64, DecoderConv64],
        [EncoderConv64Norm, DecoderConv64Norm],
        [EncoderFC, DecoderFC],
        [EncoderLinear, DecoderLinear],
    ],
)
def test_ae_models(encoder_cls: DisentEncoder, decoder_cls: DisentDecoder):
    x_shape, z_size = (3, 64, 64), 8
    # create model
    vae = AutoEncoder(
        encoder_cls(x_shape=x_shape, z_size=z_size, z_multiplier=2),
        decoder_cls(x_shape=x_shape, z_size=z_size, z_multiplier=1),
    )
    # feed forward
    with torch.no_grad():
        x = torch.randn(1, *x_shape, dtype=torch.float32)
        assert x.shape == (1, *x_shape)
        z0, z1 = vae.encode(x, chunk=True)
        assert z0.shape == (1, z_size)
        assert z1.shape == (1, z_size)
        y = vae.decode(z0)
        assert y.shape == (1, *x_shape)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import signal
import sys
from multiprocessing import current_process
from typing import List
from typing import Optional
from typing import Sequence

from lightning.pytorch import Trainer
from lightning.pytorch.loggers import Logger

from disent.util.lightning.logger_util import wb_yield_loggers

log = logging.getLogger(__name__)


# ========================================================================= #
# HYDRA CONFIG HELPERS                                                      #
# ========================================================================= #

_PL_SIGNALS_OLD_HANDLERS = {}
_PL_SIGNALS = (  # we can't capture SIGKILL
    signal.SIGINT,  # interrupted from the dialogue station
    signal.SIGTERM,  # terminate the process in a soft way
    signal.SIGABRT,  # abnormal termination
    signal.SIGSEGV,  # segmentation fault
)

_PL_LOGGERS: Optional[List[Logger]] = None
_PL_TRAINER: Optional[Trainer] = None


def safe_unset_debug_trainer():
    global _PL_TRAINER
    if _PL_TRAINER is not None:
        _PL_TRAINER = None


def set_debug_trainer(trainer: Optional[Trainer]):
    global _PL_TRAINER
    assert _PL_TRAINER is None, "debug trainer has already been set"
    _PL_TRAINER = trainer
    return trainer


def _signal_handler_log_and_exit(signal_number, frame):
    # call in all the child processes for the best chance of clearing this...
    # remove callbacks from trainer so we aren't stuck running forever!
    # TODO: this is a hack... there must be a better way to do this... could it be a pl bug?
    #       this logic is duplicated in the framework training_step
    if _PL_TRAINER and _PL_TRAINER.callbacks:
        _PL_TRAINER.callbacks.clear()

    # make sure that we only exit in the parent process
    if current_process().name != "MainProcess":
        log.debug("Skipping signal handling for child process!")
        return
    # get the signal name
    numbers_to_names = dict(
        (k, v) for v, k in reversed(sorted(signal.__dict__.items())) if v.startswith("SIG") and not v.startswith("SIG_")
    )
    signal_name = numbers_to_names.get(signal_number, signal_number)
    # log everything!
    log_error_and_exit(
        err_type=f"received exit signal",
        err_msg=f"{signal_name}",
        exit_code=signal_number,
        exc_info=False,
    )


def safe_unset_debug_loggers():
    global _PL_LOGGERS
    # unset logger
    if _PL_LOGGERS is not None:
        _PL_LOGGERS = None
        # return control to original handlers
        for signal_type in _PL_SIGNALS:
            if signal_type in _PL_SIGNALS_OLD_HANDLERS:
                handler = _PL_SIGNALS_OLD_HANDLERS.pop(signal_type)
                signal.signal(signal_type, handler)


def set_debug_loggers(loggers: Optional[Sequence[Logger]]):
    global _PL_LOGGERS
    assert _PL_LOGGERS is None, "debug logger has already been set"
    _PL_LOGGERS = loggers
    # set initial messages
    if _PL_LOGGERS is not None:
        for logger in _PL_LOGGERS:
            logger.log_metrics(
                {
                    "error_type": "N/A",
                    "error_msg": "N/A",
                    "error_occurred": False,
                }
            )
    # register signal listeners
    for signal_type in _PL_SIGNALS:
        # save the old handler
        _PL_SIGNALS_OLD_HANDLERS[signal_type] = signal.getsignal(signal_type)
        # update the handler
        signal.signal(signal_type, _signal_handler_log_and_exit)
    # return the logger
    return loggers


def log_error_and_exit(err_type: str, err_msg: str, exit_code: int = 1, exc_info=True):
    # truncate error
    err_msg = err_msg[:244] + " <TRUNCATED>" if len(err_msg) > 244 else err_msg
    # log something at least
    log.error(f"exiting: {err_type} | {err_msg}", exc_info=exc_info)
    # try log to pytorch lightning & wandb
    if _PL_LOGGERS is not None:
        for logger in _PL_LOGGERS:
            logger.log_metrics(
                {
                    "error_type": err_type,
                    "error_msg": err_msg,
                    "error_occurred": True,
                }
            )
        for wb_logger in wb_yield_loggers(_PL_LOGGERS):
            # so I dont have to scroll up... I'm lazy...
            run_url = wb_logger.experiment.get_url()
            project_url = wb_logger.experiment.get_project_url()
            log.error(f'wandb: run url: {run_url if run_url else "N/A"}')
            log.error(f'wandb: project url: {project_url if run_url else "N/A"}')
            # make sure we log everything online!
            wb_logger.experiment.finish(exit_code=exit_code)
    # EXIT!
    sys.exit(exit_code)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from numbers import Number
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import final

import torch

from disent.frameworks._ae_mixin import _AeAndVaeMixin
from disent.frameworks.helper.util import detach_all
from disent.model import AutoEncoder
from disent.util.iters import map_all

# ========================================================================= #
# framework_vae                                                             #
# ========================================================================= #


class Ae(_AeAndVaeMixin):
    """
    Basic Auto Encoder
    ------------------

    See the docs for the VAE, while the AE is more simple, the VAE docs
    cover the concepts needed to get started with writing Auto-Encoder
    sub-classes.

    Like the VAE, the AE is also written such that you can change the
    number of required input observations that should be fed through the
    network in parallel with `REQUIRED_OBS`. Various hooks are also made
    available to add functionality and access the internal data.

    - HOOKS:
        * `hook_ae_intercept_zs`
        * `hook_ae_compute_ave_aug_loss` (NB: not the same as `hook_compute_ave_aug_loss` from VAEs)

    - OVERRIDES:
        * `compute_ave_recon_loss`
    """

    # override
    REQUIRED_Z_MULTIPLIER = 1
    REQUIRED_OBS = 1

    @dataclass
    class cfg(_AeAndVaeMixin.cfg):
        pass

    def __init__(self, model: AutoEncoder, cfg: cfg = None, batch_augment=None):
        super().__init__(cfg=cfg, batch_augment=batch_augment)
        # initialise the auto-encoder mixin (recon handler, model, enc, dec, etc.)
        self._init_ae_mixin(model=model)

    # --------------------------------------------------------------------- #
    # AE Training Step -- Overridable                                       #
    # --------------------------------------------------------------------- #

    @final
    def do_training_step(self, batch, batch_idx):
        xs, xs_targ = self._get_xs_and_targs(batch, batch_idx)

        # FORWARD
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
        # latent variables
        zs = map_all(self.encode, xs)
        # [HOOK] intercept latent variables
        zs, logs_intercept_zs = self.hook_ae_intercept_zs(zs)
        # reconstruct without the final activation
        xs_partial_recon = map_all(self.decode_partial, detach_all(zs, if_=self.cfg.detach_decoder))
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

        # LOSS
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
        # compute all the recon losses
        recon_loss, logs_recon = self.compute_ave_recon_loss(xs_partial_recon, xs_targ)
        # [HOOK] augment loss
        aug_loss, logs_aug = self.hook_ae_compute_ave_aug_loss(
            zs=zs, xs_partial_recon=xs_partial_recon, xs_targ=xs_targ
        )
        # compute combined loss
        loss = 0
        if not self.cfg.disable_rec_loss:
            loss += recon_loss
        if not self.cfg.disable_aug_loss:
            loss += aug_loss
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

        # log general
        self.log_dict(
            {
                **logs_intercept_zs,
                **logs_recon,
                **logs_aug,
            }
        )

        # log progress bar
        self.log_dict(
            {
                "recon_loss": float(recon_loss),
                "aug_loss": float(aug_loss),
            },
            prog_bar=True,
        )

        # return values
        return loss

    # --------------------------------------------------------------------- #
    # Overrideable                                                          #
    # --------------------------------------------------------------------- #

    def hook_ae_intercept_zs(self, zs: Sequence[torch.Tensor]) -> Tuple[Sequence[torch.Tensor], Dict[str, Any]]:
        return zs, {}

    def hook_ae_compute_ave_aug_loss(
        self, zs: Sequence[torch.Tensor], xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        return 0, {}

    def compute_ave_recon_loss(
        self, xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        # compute reconstruction loss
        pixel_loss = self.recon_handler.compute_ave_loss_from_partial(xs_partial_recon, xs_targ)
        # return logs
        return pixel_loss, {"pixel_loss": pixel_loss}

    # --------------------------------------------------------------------- #
    # AE Model Utility Functions (Visualisation)                            #
    # --------------------------------------------------------------------- #

    @final
    def encode(self, x: torch.Tensor) -> torch.Tensor:
        """Get the deterministic latent representation (useful for visualisation)"""
        return self._model.encode(x)

    @final
    def decode(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into reconstruction x_recon (useful for visualisation)"""
        return self.recon_handler.activate(self._model.decode(z))

    @final
    def forward(self, batch: torch.Tensor) -> torch.Tensor:
        """Feed through the full deterministic model (useful for visualisation)"""
        return self.decode(self.encode(batch))

    # --------------------------------------------------------------------- #
    # AE Model Utility Functions (Training)                                 #
    # --------------------------------------------------------------------- #

    @final
    def decode_partial(self, z: torch.Tensor) -> torch.Tensor:
        """Decode latent vector z into partial reconstructions that exclude the final activation if there is one."""
        return self._model.decode(z)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
